"""Module for stacking lensing results after pre-computation."""

import numpy as np
from astropy.table import Table
from astropy.cosmology import FlatLambdaCDM
from . import surveys
from .physics import mpc_per_degree, lens_magnification_shear_bias
from .physics import critical_surface_density

__all__ = ['number_of_pairs', 'raw_tangential_shear',
           'raw_excess_surface_density', 'photo_z_dilution_factor',
           'boost_factor', 'scalar_shear_response_factor',
           'matrix_shear_response_factor', 'shear_responsivity_factor',
           'mean_lens_redshift', 'mean_source_redshift',
           'tangential_shear', 'excess_surface_density']


def number_of_pairs(table_l):
    """Compute the number of lens-source pairs per bin.

    Parameters
    ----------
    table_l : astropy.table.Table
        Precompute results for the lenses.

    Returns
    -------
    n_pairs : numpy.ndarray
        The number of lens-source pairs in each radial bin.

    """
    return np.sum(table_l['sum 1'].data, axis=0)


def raw_tangential_shear(table_l):
    """Compute the average tangential shear for a catalog.

    Parameters
    ----------
    table_l : astropy.table.Table
        Precompute results for the lenses.

    Returns
    -------
    delta_sigma : numpy.ndarray
        The raw, uncorrected tangential shear in each radial bin.

    """
    return (np.sum(table_l['sum w_ls e_t'].data *
                   table_l['w_sys'].data[:, None], axis=0) /
            np.sum(table_l['sum w_ls'].data * table_l['w_sys'].data[:, None],
                   axis=0))


def raw_excess_surface_density(table_l):
    """Compute the raw, uncorrected excess surface density for a catalog.

    Parameters
    ----------
    table_l : astropy.table.Table
        Precompute results for the lenses.

    Returns
    -------
    delta_sigma : numpy.ndarray
        The raw, uncorrected excess surface density in each radial bin.

    """
    return (np.sum(table_l['sum w_ls e_t sigma_crit'].data *
                   table_l['w_sys'].data[:, None], axis=0) /
            np.sum(table_l['sum w_ls'].data *
                   table_l['w_sys'].data[:, None], axis=0))


def photo_z_dilution_factor(table_l):
    r"""Compute the photometric redshift bias averaged over the entire catalog.

    Parameters
    ----------
    table_l : astropy.table.Table
        Precompute results for the lenses.

    Returns
    -------
    f_bias : float
        Photometric redshift bias :math:`f_{\mathrm{bias}}`.

    """
    return (np.sum(table_l['sum w_ls e_t sigma_crit f_bias'].data *
                   table_l['w_sys'].data[:, None], axis=0) /
            np.sum(table_l['sum w_ls e_t sigma_crit'].data *
                   table_l['w_sys'].data[:, None], axis=0))


def boost_factor(table_l, table_r):
    """Compute the boost factor.

    Boost factor is computed by comparing the number of lens-source pairs
    in real lenses and random lenses.

    Parameters
    ----------
    table_l : astropy.table.Table
        Precompute results for the lenses.
    table_r : astropy.table.Table, optional
        Precompute results for random lenses.

    Returns
    -------
    b : numpy.ndarray
        Boost factor in each radial bin.

    """
    return (
        np.sum(table_l['sum w_ls'].data *
               table_l['w_sys'].data[:, None], axis=0) /
        np.sum(table_l['w_sys'].data) /
        np.sum(table_r['sum w_ls'].data *
               table_r['w_sys'].data[:, None], axis=0) *
        np.sum(table_r['w_sys'].data))


def scalar_shear_response_factor(table_l):
    r"""Compute the mean shear response.

    The shear response factor :math:`m` is defined such that
    :math:`\gamma_{\mathrm obs} = (1 + m) \gamma_{\mathrm intrinsic}`.

    Parameters
    ----------
    table_l : astropy.table.Table
        Precompute results for the lenses.

    Returns
    -------
    m : numpy.ndarray
        Multiplicative shear bias in each radial bin.

    """
    return (
        np.sum(table_l['sum w_ls m'].data *
               table_l['w_sys'].data[:, None], axis=0) /
        np.sum(table_l['sum w_ls'].data *
               table_l['w_sys'].data[:, None], axis=0))


def matrix_shear_response_factor(table_l):
    r"""Compute the mean tangential response.

    The tangential shear response factor:math:`R_t` is defined such that
    :math:`\gamma_{\mathrm obs} = R_t \gamma_{\mathrm intrinsic}`.

    Parameters
    ----------
    table_l : astropy.table.Table
        Precompute results for the lenses.

    Returns
    -------
    r_t : numpy.ndarray
        Tangential shear response factor in each radial bin.

    """
    return (
        np.sum(table_l['sum w_ls R_T'] * table_l['w_sys'][:, None],
               axis=0) /
        np.sum(table_l['sum w_ls'] * table_l['w_sys'][:, None], axis=0))


def shear_responsivity_factor(table_l):
    """Compute the shear responsitivity factor.

    Parameters
    ----------
    table_l : astropy.table.Table
        Precompute results for the lenses.

    Returns
    -------
    r : numpy.ndarray
        Shear responsitivity factor in each radial bin.

    """
    return (
        np.sum(table_l['sum w_ls (1 - e_rms^2)'] *
               table_l['w_sys'][:, None], axis=0) /
        np.sum(table_l['sum w_ls'] * table_l['w_sys'][:, None], axis=0))


def mean_lens_redshift(table_l):
    """Compute the weighted-average lens redshift.

    Parameters
    ----------
    table_l : astropy.table.Table
        Precompute results for the lenses.

    Returns
    -------
    z_l : numpy.ndarray
        Mean lens redshift in each bin.

    """
    return (
        np.sum(table_l['sum w_ls z_l'] * table_l['w_sys'][:, None], axis=0) /
        np.sum(table_l['sum w_ls'] * table_l['w_sys'][:, None], axis=0))


def mean_source_redshift(table_l, photo_z_correction=False):
    """Compute the weighted-average source redshift.

    Parameters
    ----------
    table_l : astropy.table.Table
        Precompute results for the lenses.
    photo_z_correction : boolean, optional
        By default, this function returns the average photometric source
        redshift. If True and a calibration catalog or source redshift
        distribution has been provided at the precompute stage, estimate the
        intrinsic source redshift distribution. Default is False.

    Returns
    -------
    z_s : numpy.ndarray
        Mean source redshift in each bin.

    """
    if not photo_z_correction:
        key_num = 'sum w_ls z_s'
    else:
        key_num = 'sum w_ls (z_s - delta z_s)'

    return (
        np.sum(table_l[key_num] * table_l['w_sys'][:, None], axis=0) /
        np.sum(table_l['sum w_ls'] * table_l['w_sys'][:, None], axis=0))


def lens_magnification_bias(table_l, alpha_l, camb_results,
                            photo_z_correction=True):
    """Estimate the additive lens magnification bias.

    Parameters
    ----------
    table_l : astropy.table.Table
        Precompute results for the lenses.
    alpha_l : float
        TBD
    camb_results : camb.results.CAMBdata
        CAMB results object that contains information on cosmology and the
        matter power spectrum.
    photo_z_correction : boolean, optional
        Whether to correct for photo-z dilution and offsets.

    Returns
    -------
    ds_lm : numpy.ndarray
        The lens magnification bias in each radial bin.

    """
    cosmology = FlatLambdaCDM(H0=table_l.meta['H0'], Om0=table_l.meta['Om0'])

    z_l = mean_lens_redshift(table_l)
    z_s = mean_source_redshift(table_l)
    if photo_z_correction:
        z_s_true = mean_source_redshift(table_l, photo_z_correction=True)
        try:
            sigma_crit = critical_surface_density(
                z_l, z_s, cosmology, comoving=table_l.meta['comoving'])
            sigma_crit *= photo_z_dilution_factor(table_l)
        except KeyError:
            sigma_crit = critical_surface_density(
                z_l, z_s_true, cosmology, comoving=table_l.meta['comoving'])
    else:
        sigma_crit = critical_surface_density(
            z_l, z_s, cosmology, comoving=table_l.meta['comoving'])
        z_s_true = z_s

    bins = table_l.meta['bins']

    shear_mode = ('shear_mode' in table_l.meta.keys() and
                  table_l.meta['shear_mode'])

    if shear_mode:
        theta = 2.0 / 3.0 * np.diff(bins**3) / np.diff(bins**2)
    else:
        rp = 2.0 / 3.0 * np.diff(bins**3) / np.diff(bins**2)
        theta = np.deg2rad(rp / mpc_per_degree(
            z_l, cosmology=cosmology, comoving=table_l.meta['comoving']))

    gamma = np.array([lens_magnification_shear_bias(
        theta[i], alpha_l, z_l[i], z_s_true[i], camb_results) for i in
        range(len(theta))])

    if shear_mode:
        return gamma
    else:
        return gamma * sigma_crit


def tangential_shear(table_l, table_r=None, boost_correction=False,
                     scalar_shear_response_correction=False,
                     matrix_shear_response_correction=False,
                     shear_responsivity_correction=False,
                     hsc_selection_bias_correction=False,
                     random_subtraction=False, return_table=False):
    """Compute the mean tangential shear with corrections, if applicable.

    Parameters
    ----------
    table_l : astropy.table.Table
        Precompute results for the lenses.
    table_r : astropy.table.Table, optional
        Precompute results for random lenses. Default is None.
    boost_correction : boolean, optional
        If True, calculate and apply a boost factor correction. This can only
        be done if a random catalog is provided. Default is False.
    scalar_shear_response_correction : boolean or string, optional
        Whether to correct for the multiplicative shear bias (scalar form).
        Default is False.
    matrix_shear_response_correction : boolean or string, optional
        Whether to correct for the multiplicative shear bias (tensor form).
        Default is False.
    shear_responsivity_correction : boolean, optional
        If True, correct for the shear responsivity. Default is False.
    hsc_selection_bias_correction : boolean, optional
        If True, correct for the multiplicative selection bias in HSC. Default
        is False.
    random_subtraction : boolean, optional
        If True, subtract the signal around randoms. This can only be done if
        a random catalog is provided. Default is False.
    return_table : boolean, optional
        If True, return a table with many intermediate steps of the
        computation. Otherwise, a simple array with just the final tangential
        shearis returned. Default is False.

    Returns
    -------
    e_t : numpy.ndarray or astropy.table.Table
        The tangential shear in each radial bin specified in the precomputation
        phase. If `return_table` is True, will return a table with detailed
        information for each radial bin. The final result is in the column
        `et`.

    Raises
    ------
    ValueError
        If boost or random subtraction correction are requested but no random
        catalog is provided.

    """
    result = Table()

    result['rp_min'] = table_l.meta['bins'][:-1]
    result['rp_max'] = table_l.meta['bins'][1:]
    result['n_pairs'] = number_of_pairs(table_l)
    result['rp'] = np.sqrt(result['rp_min'] * result['rp_max'])
    result['et_raw'] = raw_tangential_shear(table_l)
    result['et'] = raw_tangential_shear(table_l)
    result['z_l'] = mean_lens_redshift(table_l)
    result['z_s'] = mean_source_redshift(table_l)

    if boost_correction:
        if table_r is None:
            raise ValueError('Cannot compute boost factor correction without' +
                             ' results from a random catalog.')
        result['b'] = boost_factor(table_l, table_r)
        result['et'] *= result['b']

    if scalar_shear_response_correction:
        result['1+m'] = 1 + scalar_shear_response_factor(table_l)
        result['et'] /= result['1+m']

    if matrix_shear_response_correction:
        result['R_t'] = matrix_shear_response_factor(table_l)
        result['et'] /= result['R_t']

    if shear_responsivity_correction:
        result['2R'] = 2 * shear_responsivity_factor(table_l)
        result['et'] /= result['2R']

    if hsc_selection_bias_correction:
        result['1+m_sel'] = 1 + surveys.hsc.selection_bias_factor(
            table_l)
        result['et'] *= result['1+m_sel']

    if random_subtraction:
        if table_r is None:
            raise ValueError('Cannot subtract random results without ' +
                             'results from a random catalog.')
        result['et_r'] = excess_surface_density(
            table_r, boost_correction=False,
            scalar_shear_response_correction=scalar_shear_response_correction,
            matrix_shear_response_correction=matrix_shear_response_correction,
            shear_responsivity_correction=shear_responsivity_correction,
            hsc_selection_bias_correction=hsc_selection_bias_correction,
            random_subtraction=False, return_table=False)
        result['et'] -= result['et_r']

    if not return_table:
        return result['et'].data

    return result


def excess_surface_density(table_l, table_r=None,
                           photo_z_dilution_correction=False,
                           boost_correction=False,
                           scalar_shear_response_correction=False,
                           matrix_shear_response_correction=False,
                           shear_responsivity_correction=False,
                           hsc_selection_bias_correction=False,
                           random_subtraction=False,
                           return_table=False):
    """Compute the mean excess surface density with corrections, if applicable.

    Parameters
    ----------
    table_l : astropy.table.Table
        Precompute results for the lenses.
    table_r : astropy.table.Table, optional
        Precompute results for random lenses. Default is None.
    photo_z_dilution_correction : boolean, optional
        If True, correct for photo-z biases. This can only be done if a
        calibration catalog has been provided in the Precomputation phase.
        Default is False.
    boost_correction : boolean, optional
        If true, calculate and apply a boost factor correction. This can only
        be done if a random catalog is provided. Default is False.
    scalar_shear_response_correction : boolean or string, optional
        Whether to correct for the multiplicative shear bias (scalar form).
        Default is False.
    matrix_shear_response_correction : boolean or string, optional
        Whether to correct for the multiplicative shear bias (tensor form).
        Default is False.
    shear_responsivity_correction : boolean, optional
        If True, correct for the shear responsivity. Default is False.
    hsc_selection_bias_correction : boolean, optional
        If True, correct for the multiplicative selection bias in HSC. Default
        is False.
    random_subtraction : boolean, optional
        If True, subtract the signal around randoms. This can only be done if
        a random catalog is provided. Default is False.
    return_table : boolean, optional
        If True, return a table with many intermediate steps of the
        computation. Otherwise, a simple array with just the final excess
        surface density is returned. Default is False.

    Returns
    -------
    delta_sigma : numpy.ndarray or astropy.table.Table
        The excess surface density in each radial bin specified in the
        precomputation phase. If `return_table` is True, will return a table
        with detailed information for each radial bin. The final result is in
        the column `ds`.

    Raises
    ------
    ValueError
        If boost or random subtraction correction are requested but no random
        catalog is provided.

    """
    result = Table()

    result['rp_min'] = table_l.meta['bins'][:-1]
    result['rp_max'] = table_l.meta['bins'][1:]
    result['n_pairs'] = number_of_pairs(table_l)
    result['rp'] = np.sqrt(result['rp_min'] * result['rp_max'])
    result['ds_raw'] = raw_excess_surface_density(table_l)
    result['ds'] = raw_excess_surface_density(table_l)
    result['z_l'] = mean_lens_redshift(table_l)
    result['z_s'] = mean_source_redshift(table_l)

    if boost_correction:
        if table_r is None:
            raise ValueError('Cannot compute boost factor correction without' +
                             ' results from a random catalog.')
        result['b'] = boost_factor(table_l, table_r)
        result['ds'] *= result['b']

    if scalar_shear_response_correction:
        result['1+m'] = 1 + scalar_shear_response_factor(table_l)
        result['ds'] /= result['1+m']

    if matrix_shear_response_correction:
        result['R_t'] = matrix_shear_response_factor(table_l)
        result['ds'] /= result['R_t']

    if shear_responsivity_correction:
        result['2R'] = 2 * shear_responsivity_factor(table_l)
        result['ds'] /= result['2R']

    if hsc_selection_bias_correction:
        result['1+m_sel'] = 1 + surveys.hsc.selection_bias_factor(
            table_l)
        result['ds'] *= result['1+m_sel']

    if photo_z_dilution_correction:
        result['f_bias'] = photo_z_dilution_factor(table_l)
        result['ds'] *= result['f_bias']

    if random_subtraction:
        if table_r is None:
            raise ValueError('Cannot subtract random results without ' +
                             'results from a random catalog.')
        result['ds_r'] = excess_surface_density(
            table_r, photo_z_dilution_correction=photo_z_dilution_correction,
            boost_correction=False,
            scalar_shear_response_correction=scalar_shear_response_correction,
            matrix_shear_response_correction=matrix_shear_response_correction,
            shear_responsivity_correction=shear_responsivity_correction,
            hsc_selection_bias_correction=hsc_selection_bias_correction,
            random_subtraction=False, return_table=False)
        result['ds'] -= result['ds_r']

    if not return_table:
        return result['ds'].data

    return result



"""Module with functions specific to the Hyper-Suprime Cam survey."""

import numpy as np

__all__ = ['default_version', 'known_versions', 'e_2_convention',
           'default_column_keys', 'apply_photo_z_quality_cut',
           'selection_bias_factor']

default_version = 'PDR2'
known_versions = ['PDR2', ]
e_2_convention = 'flipped'


def default_column_keys(version=default_version):
    """Return a dictionary of default column keys.

    Parameters
    ----------
    version : string or None, optional
        Version of the catalog.

    Returns
    -------
    keys : dict
        Dictionary of default column keys.

    Raises
    ------
    ValueError
        If `version` does not correspond to a known catalog version.

    """
    if version == 'PDR2':
        keys = {
            'ra': 'ira',
            'dec': 'idec',
            'z': 'photoz_best',
            'z_low': 'photoz_err68_min',
            'e_1': 'ishape_hsm_regauss_e1',
            'e_2': 'ishape_hsm_regauss_e2',
            'w': 'ishape_hsm_regauss_derived_shape_weight',
            'm': 'ishape_hsm_regauss_derived_shear_bias_m',
            'e_rms': 'ishape_hsm_regauss_derived_rms_e',
            'R_2': 'ishape_hsm_regauss_resolution'}
    else:
        raise ValueError(
            "Unkown version of DES. Supported versions are {}.".format(
                known_versions))

    return keys


def apply_photo_z_quality_cut(table_s, global_photo_z_cuts,
                              specinfo_photo_z_cuts):
    """Apply HSC-specific photo-z cuts to the source catalog.

    Parameters
    ----------
    table_s : astropy.table.Table
        HSC weak lensing source catalog.
    global_photo_z_cuts : string
        Requirements for global photometric redshift quality.
    specinfo_photo_z_cuts : string
        Specinfo photo-z cuts.

    Returns
    -------
    table_s : astropy.table.Table
        Table containing only source passing the cuts.

    Raises
    ------
    ValueError
        If invalid options are passed.

    """
    if global_photo_z_cuts == "basic":
        # Implement ~2-sigma clipping over chi^2_5.
        mask = table_s['frankenz_model_llmin'] < 6.
    elif global_photo_z_cuts == "medium":
        mask = table_s['frankenz_model_llmin'] < 6.
        # Remove sources with overly broad PDFs around `z_best`.
        mask = mask & (table_s['frankenz_photoz_risk_best'] < 0.25)
    elif global_photo_z_cuts == "strict":
        # Similar to `medium`, but stricter.
        mask = table_s['frankenz_model_llmin'] < 6.
        mask = mask & (table_s['frankenz_photoz_risk_best'] < 0.15)
    elif global_photo_z_cuts == "none":
        # No global photo-z cut is applied.
        mask = np.isfinite(table_s['frankenz_photoz_best'])
    else:
        raise ValueError("Invalid global photo-z cuts option "
                         "{}".format(global_photo_z_cuts))

    # Apply specinfo photo-z cuts
    # Should most likely be applied with corresponding redshift cuts.
    if specinfo_photo_z_cuts == "none":
        # No cut
        pass
    elif specinfo_photo_z_cuts == "great":
        # >50% of info comes from non-photo-z sources.
        mask = mask & (table_s['frankenz_model_ptype2'] < 0.5)
    elif specinfo_photo_z_cuts == "good":
        # >10% of info comes from non-photo-z sources.
        mask = mask & (table_s['frankenz_model_ptype2'] < 0.9)
    elif specinfo_photo_z_cuts == "moderate":
        # 10%-50% of info comes from non-photo-z sources.
        mask = mask & (table_s['frankenz_model_ptype2'] >= 0.5)
        mask = mask & (table_s['frankenz_model_ptype2'] < 0.9)
    elif specinfo_photo_z_cuts == "poor":
        # <=50% of info comes from non-photo-z sources.
        mask = mask & (table_s['frankenz_model_ptype2'] >= 0.5)
    elif specinfo_photo_z_cuts == "poorest":
        # <=10% of info comes from non-photo-z sources.
        mask = mask & (table_s['frankenz_model_ptype2'] >= 0.9)
    else:
        raise ValueError("Invalid specinfo photo-z cuts option "
                         "= {}".format(global_photo_z_cuts))

    return table_s[mask]


def selection_bias_factor(table_l):
    """Compute the multiplicative selection bias.

    Parameters
    ----------
    table_l : astropy.table.Table
        Precompute results for the lenses.

    Returns
    -------
    m_sel : numpy.ndarray
        Multiplicative selection bias in each radial bin.

    """
    return (
        np.sum(table_l['sum w_ls A p(R_2=0.3)'] *
               table_l['w_sys'][:, None], axis=0) /
        np.sum(table_l['sum w_ls'] * table_l['w_sys'][:, None], axis=0))



"""Delta sigma pipeline for galaxy-galaxy lensing."""

from . import helpers
from . import jackknife
from . import physics
from . import precompute
from . import stacking
from . import surveys
import importlib.metadata


__all__ = ["helpers", "jackknife", "physics", "precompute", "stacking",
           "surveys"]
__version__ = importlib.metadata.version("dsigma")



# cython: language_level=3, boundscheck=False, wraparound=False
# cython: nonecheck=False, cdivision=True, initializedcheck=False

import queue as Queue
import numpy as np
from tqdm import tqdm
from astropy_healpix import HEALPix
from astropy import units as u
from libc.math cimport sin, cos, sqrt, fmax, pow
from scipy.spatial import cKDTree

from astropy import constants as c
from astropy import units as u


cdef double sigma_crit_factor = (
    1e-6 * c.c**2 / (4 * np.pi * c.G)).to(u.Msun / u.pc).value
cdef double deg2rad = np.pi / 180.0

cdef double dx, dy, dz

cdef double dist_3d_sq(double sin_ra_1, double cos_ra_1, double sin_dec_1,
                       double cos_dec_1, double sin_ra_2, double cos_ra_2,
                       double sin_dec_2, double cos_dec_2):

    dx = cos_ra_1 * cos_dec_1 - cos_ra_2 * cos_dec_2
    dy = sin_ra_1 * cos_dec_1 - sin_ra_2 * cos_dec_2
    dz = sin_dec_1 - sin_dec_2

    return dx * dx + dy * dy + dz * dz


def precompute_engine(
        u_pix_l, n_pix_l_in, u_pix_s, n_pix_s_in, dist_3d_sq_bins_in,
        table_l, table_s, table_r, bins, bint comoving, float weighting,
        int nside, queue, progress_bar):

    cdef long[::1] n_pix_l = n_pix_l_in
    cdef long[::1] n_pix_s = n_pix_s_in

    cdef double[::1] z_l = table_l['z']
    cdef double[::1] z_s = table_s['z']
    cdef double[::1] d_com_l = table_l['d_com']
    cdef double[::1] d_com_s = table_s['d_com']
    cdef double[::1] sin_ra_l = table_l['sin ra']
    cdef double[::1] cos_ra_l = table_l['cos ra']
    cdef double[::1] sin_dec_l = table_l['sin dec']
    cdef double[::1] cos_dec_l = table_l['cos dec']
    cdef double[::1] sin_ra_s = table_s['sin ra']
    cdef double[::1] cos_ra_s = table_s['cos ra']
    cdef double[::1] sin_dec_s = table_s['sin dec']
    cdef double[::1] cos_dec_s = table_s['cos dec']
    cdef double[::1] w_s = table_s['w']
    cdef double[::1] e_1 = table_s['e_1']
    cdef double[::1] e_2 = table_s['e_2']
    cdef double[::1] z_l_max = table_s['z_l_max']

    cdef bint has_sigma_crit_eff = 'sigma_crit_eff' in table_l.keys()
    cdef int n_z_bins = 0
    cdef double[::1] sigma_crit_eff
    cdef long[::1] z_bin
    if has_sigma_crit_eff:
        n_z_bins = len(table_l['sigma_crit_eff']) // len(table_l['z'])
        sigma_crit_eff = table_l['sigma_crit_eff']
        z_bin = table_s['z_bin']

    cdef bint has_m = 'm' in table_s.keys()
    cdef double[::1] m
    if has_m:
        m = table_s['m']

    cdef bint has_e_rms = 'e_rms' in table_s.keys()
    cdef double[::1] e_rms
    if has_e_rms:
        e_rms = table_s['e_rms']

    cdef bint has_R_2 = 'R_2' in table_s.keys()
    cdef double[::1] R_2
    if has_R_2:
        R_2 = table_s['R_2']

    cdef bint has_R_matrix = 'R_11' in table_s.keys()
    cdef double[::1] R_11, R_12, R_21, R_22
    if has_R_matrix:
        R_11 = table_s['R_11']
        R_12 = table_s['R_12']
        R_21 = table_s['R_21']
        R_22 = table_s['R_22']

    cdef double[::1] dist_3d_sq_bins = dist_3d_sq_bins_in

    cdef long[::1] sum_1 = table_r['sum 1']
    cdef double[::1] sum_w_ls = table_r['sum w_ls']
    cdef double[::1] sum_w_ls_e_t = table_r['sum w_ls e_t']
    cdef double[::1] sum_w_ls_e_t_sigma_crit = table_r['sum w_ls e_t sigma_crit']
    cdef double[::1] sum_w_ls_z_s = table_r['sum w_ls z_s']
    cdef double[::1] sum_w_ls_m
    if has_m:
        sum_w_ls_m = table_r['sum w_ls m']
    cdef double[::1] sum_w_ls_1_minus_e_rms_sq
    if has_e_rms:
        sum_w_ls_1_minus_e_rms_sq = table_r['sum w_ls (1 - e_rms^2)']
    cdef double[::1] sum_w_ls_A_p_R_2
    if has_R_2:
        sum_w_ls_A_p_R_2 = table_r['sum w_ls A p(R_2=0.3)']
    cdef double[::1] sum_w_ls_R_T
    if has_R_matrix:
        sum_w_ls_R_T = table_r['sum w_ls R_T']

    hp = HEALPix(nside, order='ring')
    lon, lat = hp.healpix_to_lonlat(np.arange(hp.npix))
    x = np.cos(lon) * np.cos(lat)
    y = np.sin(lon) * np.cos(lat)
    z = np.sin(lat)
    xyz = np.array([x, y, z]).T
    xyz_l = xyz[u_pix_l]
    xyz_s = xyz[u_pix_s]
    kdtree = cKDTree(xyz_s)

    cdef long pix_l, i_l, i_l_min, i_l_max
    cdef long pix_s, i_pix_s, l_pix_s, i_s, i_s_min, i_s_max
    cdef long[::1] pix_s_list
    cdef long i_bin, n_bins = len(bins) - 1
    cdef long offset_bin, offset_result
    cdef double dist_3d_sq_max, dist_3d_sq_ls
    cdef double sin_ra_l_minus_ra_s, cos_ra_l_minus_ra_s
    cdef double sin_2phi, cos_2phi, tan_phi, tan_phi_num, tan_phi_den, e_t
    cdef double w_ls, sigma_crit
    cdef double max_pixrad = 1.05 * hp.pixel_resolution.to(u.deg).value
    cdef double inf = float('inf'), summand

    if progress_bar:
        pbar = tqdm(total=len(u_pix_l))

    while True:

        # Check whether there is still a lens pixel in the queue.
        try:
            pix_l = queue.get(timeout=0.5)
        except Queue.Empty:
            break

        if pix_l == 0:
            i_l_min = 0
        else:
            i_l_min = n_pix_l[pix_l - 1]
        i_l_max = n_pix_l[pix_l]

        # Find the maximum angular search radius.
        dist_3d_sq_max = 0.0
        for i_l in range(i_l_min, i_l_max):
            dist_3d_sq_max = fmax(dist_3d_sq_bins[i_l * (n_bins + 1) + n_bins],
                                  dist_3d_sq_max)
        # Note that pixels can be up to max_pixrad away from the pixel center.
        dist_3d_sq_max += (4 * deg2rad * deg2rad * max_pixrad * max_pixrad +
                           4 * sqrt(dist_3d_sq_max) * deg2rad * max_pixrad)

        # Get list of all source pixels that could contain suitable sources.
        pix_s_list = np.fromiter(
            kdtree.query_ball_point(xyz_l[pix_l], sqrt(dist_3d_sq_max)),
            dtype=long)
        l_pix_s = len(pix_s_list)

        # Loop over all suitable source pixels.
        for i_pix_s in range(l_pix_s):

            pix_s = pix_s_list[i_pix_s]
            if pix_s == 0:
                i_s_min = 0
            else:
                i_s_min = n_pix_s[pix_s - 1]
            i_s_max = n_pix_s[pix_s]

            # Loop over all lenses in the pixel.
            for i_l in range(i_l_min, i_l_max):

                offset_result = i_l * n_bins
                offset_bin = i_l * (n_bins + 1)

                # Loop over all sources in the pixel.
                for i_s in range(i_s_min, i_s_max):

                    if z_l[i_l] > z_l_max[i_s]:
                        continue

                    dist_3d_sq_ls = dist_3d_sq(
                        sin_ra_l[i_l], cos_ra_l[i_l], sin_dec_l[i_l],
                        cos_dec_l[i_l], sin_ra_s[i_s], cos_ra_s[i_s],
                        sin_dec_s[i_s], cos_dec_s[i_s])

                    i_bin = n_bins
                    while i_bin >= 0:
                        if dist_3d_sq_ls > dist_3d_sq_bins[offset_bin + i_bin]:
                            break
                        i_bin -= 1

                    if i_bin == n_bins or i_bin < 0:
                        continue

                    if w_s[i_s] == 0:
                        continue

                    if has_sigma_crit_eff:
                        sigma_crit = sigma_crit_eff[
                            i_l * n_z_bins + z_bin[i_s]]
                    elif z_l[i_l] < z_s[i_s]:
                        sigma_crit = (sigma_crit_factor * (1 + z_l[i_l]) *
                            d_com_s[i_s] / d_com_l[i_l] /
                            (d_com_s[i_s] - d_com_l[i_l]))
                        if comoving:
                            sigma_crit /= (1.0 + z_l[i_l]) * (1.0 + z_l[i_l])
                    else:
                        if weighting < 0:
                            continue
                        else:
                            sigma_crit = inf

                    if weighting == 0:
                        w_ls = w_s[i_s]
                    elif weighting == -2:
                         w_ls = w_s[i_s] / sigma_crit / sigma_crit
                    else:
                        w_ls = w_s[i_s] * pow(sigma_crit, weighting)

                    if w_ls == 0:
                        continue

                    sin_ra_l_minus_ra_s = (sin_ra_l[i_l] * cos_ra_s[i_s] -
                                           cos_ra_l[i_l] * sin_ra_s[i_s])
                    cos_ra_l_minus_ra_s = (cos_ra_l[i_l] * cos_ra_s[i_s] +
                                           sin_ra_l[i_l] * sin_ra_s[i_s])
                    tan_phi_num = (cos_dec_s[i_s] * sin_dec_l[i_l] - sin_dec_s[i_s] *
                                   cos_dec_l[i_l] * cos_ra_l_minus_ra_s)
                    tan_phi_den = cos_dec_l[i_l] * sin_ra_l_minus_ra_s
                    if tan_phi_den == 0:
                        cos_2phi = -1
                        sin_2phi = 0
                    else:
                        tan_phi = tan_phi_num / tan_phi_den
                        cos_2phi = (2.0 / (1.0 + tan_phi * tan_phi)) - 1.0
                        sin_2phi = 2.0 * tan_phi / (1.0 + tan_phi * tan_phi)

                    e_t = - e_1[i_s] * cos_2phi + e_2[i_s] * sin_2phi

                    sum_1[offset_result + i_bin] += 1
                    summand = w_ls
                    sum_w_ls[offset_result + i_bin] += summand
                    summand *= e_t
                    sum_w_ls_e_t[offset_result + i_bin] += summand
                    summand *= sigma_crit
                    sum_w_ls_e_t_sigma_crit[offset_result + i_bin] += summand
                    sum_w_ls_z_s[offset_result + i_bin] += w_ls * z_s[i_s]
                    if has_m:
                        sum_w_ls_m[offset_result + i_bin] += w_ls * m[i_s]
                    if has_e_rms:
                        sum_w_ls_1_minus_e_rms_sq[offset_result + i_bin] += (
                            w_ls * (1 - e_rms[i_s] * e_rms[i_s]))
                    if has_R_2 and R_2[i_s] <= 0.31:
                        sum_w_ls_A_p_R_2[offset_result + i_bin] += (
                            0.00865 * w_ls / 0.01)
                    if has_R_matrix:
                        sum_w_ls_R_T[offset_result + i_bin] += w_ls * (
                            R_11[i_s] * cos_2phi * cos_2phi +
                            R_22[i_s] * sin_2phi * sin_2phi +
                            (R_12[i_s] + R_21[i_s]) * sin_2phi * cos_2phi)

        if progress_bar:
            pbar.update(pix_l + 1 - pbar.n)

    if progress_bar:
        pbar.update(len(u_pix_l) - pbar.n)
        pbar.close()

    return 0



"""Module containing jackknife resampling functions."""

import warnings
import numpy as np
import astropy.units as u
from astropy.coordinates import SkyCoord
from sklearn.cluster import DBSCAN, MiniBatchKMeans
from scipy.spatial import cKDTree
from astropy.table import Table
from scipy.ndimage import gaussian_filter

from .helpers import spherical_to_cartesian


__all__ = ["compute_jackknife_fields", "compress_jackknife_fields",
           "smooth_correlation_matrix", "jackknife_resampling"]


def compute_jackknife_fields(table, centers, distance_threshold=1,
                             weights=None):
    """Compute the centers for jackknife regions using DBSCAN and KMeans.

    The function first runs DBSCAN to identify continous fields of points.
    Afterwards, KMeans clustering is run. The initial cluster centers are
    random points from each continous field. The number of initial cluster
    centers per field is determined according to the total weight of each
    continous field. The centers are defined in cartesian coordinates on a unit
    sphere.

    Parameters
    ----------
    table : astropy.table.Table
        Catalog containing objects. The catalog needs to have coordinates and
        field IDs.
    centers : int or numpy.ndarray
        If int, total number of jackknife fields. Otherwise, the centers
        returned from a previous call to that function. This allows for
        different samples to have the same jackknife fields.
    distance_threshold : float, optional
        The angular separation in degrees used to link points and calculate
        continous fields before running KMeans. Default is 1.
    weights : None or numpy.ndarray
        Per-lens weights for clustering. If None, assume the same weight for
        all points. Default is None.

    Returns
    -------
    centers : numpy.ndarray
        The coordinates of the centers of the jackknife regions.

    """
    x, y, z = spherical_to_cartesian(table['ra'].data, table['dec'].data)
    xyz = np.column_stack((x, y, z))
    xyz = np.column_stack(spherical_to_cartesian(
        table['ra'].data, table['dec'].data))

    if isinstance(centers, np.ndarray):
        kdtree = cKDTree(centers)
        table['field_jk'] = kdtree.query(xyz)[1]
        return centers

    if weights is None:
        weights = np.ones(len(table))

    n_jk = centers

    if not isinstance(distance_threshold, u.quantity.Quantity):
        distance_threshold *= u.deg

    eps = np.sqrt(
        2 - 2 * np.cos(distance_threshold.to(u.rad).value))
    c = DBSCAN(eps=eps, algorithm='kd_tree').fit(xyz).labels_

    w_c = np.bincount(c[c != -1], weights=weights[c != -1])
    if n_jk < len(w_c):
        raise RuntimeError(
            "The number of jackknife regions cannot be smaller than the " +
            "number of continous fields. Try increasing `distance_threshold`" +
            " or decreasing `centers`.")

    # Assign the number of jackknife fields according to the total number of
    # objects in each field.
    n_jk_per_c = np.diff(np.rint(
        np.cumsum(w_c) / np.sum(w_c) * n_jk).astype(np.int), prepend=0)

    # It can happen that one field is assigned 0 jackknife fields. In this
    # case, we will assign 1.
    while np.any(w_c[n_jk_per_c == 0] > 0):
        n_jk_per_c[np.argmin(n_jk_per_c)] += 1
        n_jk_per_c[np.argmax(n_jk_per_c)] -= 1

    init = np.zeros((0, 3))
    for i in range(len(w_c)):
        mask = i != c
        if w_c[i] > 0:
            init = np.vstack([init, xyz[~mask][np.random.choice(
                np.sum(~mask), n_jk_per_c[i], replace=False,
                p=weights[~mask] / w_c[i])]])

    centers = MiniBatchKMeans(n_clusters=n_jk, init=init, n_init=1).fit(
        xyz[weights > 0], sample_weight=weights[weights > 0]).cluster_centers_
    compute_jackknife_fields(table, centers)

    return centers


def compress_jackknife_fields(table):
    """Sum together all lenses in each jackknife field.

    After assigning jackknife fields, for most applications, we do not need
    information on individual objects anymore. Compress the information in each
    jackknife field by taking weighted averages. The only exception is the
    weight column where the sum is taken.

    Parameters
    ----------
    table : astropy.table.Table
        Catalog containing objects. The catalog needs to have been assigned
        jackknife fields.

    Returns
    -------
    table_jk : astropy.table.Table
        Catalog containing the information for each jackknife field. It has
        exactly as many rows as there are jackknife fields.

    """
    all_field_jk = np.unique(table['field_jk'])
    table_jk = Table(table[:len(all_field_jk)], copy=True)

    for i, field_jk in enumerate(all_field_jk):
        mask = table['field_jk'] == field_jk
        for key in table.colnames:
            if key == 'field_jk':
                table_jk[i][key] = table[key][mask][0]
            elif key in ['w_sys', 'sum 1']:
                table_jk[i][key] = np.sum(table[key][mask], axis=0)
            else:
                with warnings.catch_warnings():
                    if np.any(np.isnan(table[key][mask])):
                        warnings.simplefilter(
                            'ignore', category=RuntimeWarning)
                    table_jk[i][key] = np.average(
                        table[key][mask], weights=table['w_sys'][mask], axis=0)

    return table_jk


def smooth_correlation_matrix(cor, sigma, exclude_diagonal=True):
    """Apply a simple gaussian filter on a correlation matrix.

    Parameters
    ----------
    cor : numpy.ndarray
        Correlation matrix.
    sigma : int
        Scale of the gaussian filter.
    exclude_diagonal : boolean, optional
        Whether to exclude the diagonal from the smoothing. That is what should
        be done generally because the diagonal is 1 by definition. Default is
        True.

    Returns
    -------
    cor_new : numpy.ndarray
        Smoothed correlation matrix.

    """
    n_dim = len(np.diag(cor))
    cor_new = np.copy(cor)

    if exclude_diagonal:
        cor_new[0, 0] = 0.5 * (cor[0, 1] + cor[1, 0])
        cor_new[n_dim - 1, n_dim - 1] = 0.5 * (cor[n_dim - 1, n_dim - 2] +
                                               cor[n_dim - 2, n_dim - 1])

        for i in range(1, n_dim - 1):
            cor_new[i, i] = 0.25 * (cor[i, i - 1] + cor[i, i + 1] +
                                    cor[i - 1, i] + cor[i + 1, i])

    cor_new = gaussian_filter(cor_new, sigma, mode='nearest')

    if exclude_diagonal:
        for i in range(n_dim):
            cor_new[i, i] = cor[i, i]

    return cor_new


def jackknife_resampling(f, table_l, table_r=None, table_l_2=None,
                         table_r_2=None, **kwargs):
    """Compute the covariance of a function from jackknife re-sampling.

    Parameters
    ----------
    f : function
        Function that returns a result for which we want to have uncertainties.
        The function must take exactly one positional argument, the lens table.
        Additionally, it can have several additional keyword arguments.
    table_l : astropy.table.Table
        Precompute results for the lenses. The catalog must have jackknife
        regions assigned to it.
    table_r : optional, astropy.table.Table, optional
        Precompute results for random lenses. The input function must accept
        the random lens table via the `table_r` keyword argument. Default
        is None.
    table_l_2 : optional, astropy.table.Table
        Precompute results for a second set of lenses.The input function must
        accept the second lens table via the `table_l_2` keyword argument.
        Default is None.
    table_r_2 : optional, astropy.table.Table, optional
        Precompute results for a second set of random lenses. The input
        function must accept the second random lens table via the `table_r_2`
        keyword argument. Default is None.
    kwargs : dict, optional
        Additional keyword arguments to be passed to the function.

    Returns
    -------
    cov : numpy.ndarray
        Covariance matrix of the result derived from jackknife re-sampling.

    """
    samples = []

    for field_jk in np.unique(table_l['field_jk']):

        mask_l = table_l['field_jk'] != field_jk

        for name, table in zip(['table_r', 'table_l_2', 'table_r_2'],
                               [table_r, table_l_2, table_r_2]):
            if table is not None:
                kwargs[name] = table[table['field_jk'] != field_jk]

        samples.append(f(table_l[mask_l], **kwargs))

    return ((len(np.unique(table_l['field_jk'])) - 1) *
            np.cov(np.array(samples), rowvar=False, ddof=0))



"""Module for pre-computing lensing results."""

import numbers
import warnings
import multiprocessing as mp
import queue as Queue

import numpy as np
from astropy_healpix import HEALPix
from scipy.interpolate import interp1d

from astropy.cosmology import FlatLambdaCDM
from astropy import units as u
from astropy.units import UnitConversionError

from .physics import critical_surface_density
from .physics import effective_critical_surface_density
from .precompute_engine import precompute_engine


__all__ = ["photo_z_dilution_factor", "mean_photo_z_offset", "precompute"]


def photo_z_dilution_factor(z_l, table_c, cosmology, weighting=-2,
                            lens_source_cut=0):
    """Calculate the photo-z delta sigma bias as a function of lens redshift.

    Parameters
    ----------
    z_l : float or numpy.ndarray
        Redshift(s) of the lens.
    table_c : astropy.table.Table
        Photometric redshift calibration catalog.
    cosmology : astropy.cosmology
        Cosmology to assume for calculations.
    weighting : float, optional
        The exponent of weighting of each lens-source pair by the critical
        surface density. A natural choice is -2 which minimizes shape noise.
        Default is -2.
    lens_source_cut : None, float or numpy.ndarray, optional
        Determine the lens-source redshift separation cut. If None, no cut is
        applied. If a float, determines the minimum redshift separation between
        lens and source redshift for lens-source pairs to be used. If an array,
        it has to be the same length as the source table and determines the
        maximum lens redshift for a lens-source pair to be used. Default is 0.

    Returns
    -------
    f_bias : float or numpy.ndarray
        The photo-z bias factor, `f_bias`, for the lens redshift(s).

    """
    if lens_source_cut is None:
        z_l_max = np.repeat(np.amax(z_l) + 1, len(table_c))
    elif isinstance(lens_source_cut, numbers.Number):
        z_l_max = table_c['z'] - lens_source_cut
    else:
        z_l_max = np.array(lens_source_cut)

    z_s = table_c['z']
    z_s_true = table_c['z_true']
    d_l = cosmology.comoving_transverse_distance(z_l).to(u.Mpc).value
    d_s = cosmology.comoving_transverse_distance(table_c['z']).to(u.Mpc).value
    d_s_true = cosmology.comoving_transverse_distance(
        table_c['z_true']).to(u.Mpc).value
    w = table_c['w_sys'] * table_c['w']

    if hasattr(z_l, '__len__'):
        shape = (len(z_l), len(table_c))
        z_s = np.tile(z_s, len(z_l)).reshape(shape)
        z_s_true = np.tile(z_s_true, len(z_l)).reshape(shape)
        d_s = np.tile(d_s, len(z_l)).reshape(shape)
        d_s_true = np.tile(d_s_true, len(z_l)).reshape(shape)
        z_l_max = np.tile(z_l_max, len(z_l)).reshape(shape)
        w = np.tile(w, len(z_l)).reshape(shape)
        z_l = np.repeat(z_l, len(table_c)).reshape(shape)
        d_l = np.repeat(d_l, len(table_c)).reshape(shape)

    sigma_crit_phot = critical_surface_density(z_l, z_s, d_l=d_l, d_s=d_s)
    sigma_crit_true = critical_surface_density(z_l, z_s_true, d_l=d_l,
                                               d_s=d_s_true)

    mask = (z_l_max < z_l) | (z_l > z_s)

    if np.any(np.all(mask, axis=-1)):
        warnings.warn('Could not find valid calibration sources for some ' +
                      'lens redshifts. The f_bias correction may be ' +
                      'undefined.', RuntimeWarning)

    return (np.sum((w * sigma_crit_phot**weighting) * (~mask), axis=-1) /
            np.sum((w * sigma_crit_phot**(weighting + 1) / sigma_crit_true) *
                   (~mask), axis=-1))


def mean_photo_z_offset(z_l, table_c, cosmology, weighting=-2,
                        lens_source_cut=0):
    """Calculate the mean offset of source photometric redshifts.

    Parameters
    ----------
    z_l : float or numpy.ndarray
        Redshift(s) of the lens.
    table_c : astropy.table.Table, optional
        Photometric redshift calibration catalog.
    cosmology : astropy.cosmology
        Cosmology to assume for calculations.
    weighting : float, optional
        The exponent of weighting of each lens-source pair by the critical
        surface density. A natural choice is -2 which minimizes shape noise.
        Default is -2.
    lens_source_cut : None, float or numpy.ndarray, optional
        Determine the lens-source redshift separation cut. If None, no cut is
        applied. If a float, determines the minimum redshift separation between
        lens and source redshift for lens-source pairs to be used. If an array,
        it has to be the same length as the source table and determines the
        maximum lens redshift for a lens-source pair to be used. Default is 0.

    Returns
    -------
    dz : float or numpy.ndarray
        The mean source redshift offset for the lens redshift(s).

    """
    if lens_source_cut is None:
        z_l_max = np.repeat(np.amax(z_l) + 1, len(table_c))
    elif isinstance(lens_source_cut, numbers.Number):
        z_l_max = table_c['z'] - lens_source_cut
    else:
        z_l_max = np.array(lens_source_cut)

    z_s = table_c['z']
    z_s_true = table_c['z_true']
    d_l = cosmology.comoving_transverse_distance(z_l).to(u.Mpc).value
    d_s = cosmology.comoving_transverse_distance(table_c['z']).to(
        u.Mpc).value
    w = table_c['w_sys'] * table_c['w']

    if not np.isscalar(z_l):
        shape = (len(z_l), len(table_c))
        z_s = np.tile(z_s, len(z_l)).reshape(shape)
        z_s_true = np.tile(z_s_true, len(z_l)).reshape(shape)
        d_s = np.tile(d_s, len(z_l)).reshape(shape)
        z_l_max = np.tile(z_l_max, len(z_l)).reshape(shape)
        w = np.tile(w, len(z_l)).reshape(shape)
        z_l = np.repeat(z_l, len(table_c)).reshape(shape)
        d_l = np.repeat(d_l, len(table_c)).reshape(shape)

    sigma_crit = critical_surface_density(z_l, z_s, d_l=d_l, d_s=d_s)
    w = w * sigma_crit**weighting

    mask = (z_l_max < z_l) | (z_l > z_s)

    return np.sum((z_s - z_s_true) * w * (~mask), axis=-1) / np.sum(
        w * (~mask), axis=-1)


def get_raw_multiprocessing_array(array):
    """Convert a numpy array into a shared-memory multiprocessing array.

    Parameters
    ----------
    array : numpy.ndarray or None
        Input array.

    Returns
    -------
    array_mp : multiprocessing.RawArray or None
        Output array. None if input is None.

    """
    if array is None:
        return None

    array_mp = mp.RawArray('l' if np.issubdtype(array.dtype, np.integer) else
                           'd', len(array))
    array_np = np.ctypeslib.as_array(array_mp)
    array_np[:] = array

    return array_mp


def precompute(
        table_l, table_s, bins, table_c=None, table_n=None,
        cosmology=FlatLambdaCDM(H0=100, Om0=0.3), comoving=True,
        weighting=-2, lens_source_cut=0, nside=256, n_jobs=1,
        progress_bar=False):
    """For all lenses in the catalog, precompute the lensing statistics.

    Parameters
    ----------
    table_l : astropy.table.Table
        Catalog of lenses.
    table_s : astropy.table.Table
        Catalog of sources.
    bins : numpy.ndarray or u.quantity.Quantity
        Bins in radius to use for the stacking. If a numpy array, bins are
        assumed to be in Mpc. If an astropy quantity, one can pass both length
        units, e.g. kpc and Mpc, as well as angular units, i.e. deg and rad.
    table_c : astropy.table.Table, optional
        Additional photometric redshift calibration catalog. Default is None.
    table_n : astropy.table.Table, optional
        Source redshift distributions. Default is None.
    cosmology : astropy.cosmology, optional
        Cosmology to assume for calculations. Default is a flat LambdaCDM
        cosmology with h=1 and Om0=0.3.
    comoving : boolean, optional
        Whether to use comoving or physical quantities for radial bins (if
        given in physical units) and the excess surface density. Default is
        True.
    weighting : float, optional
        The exponent of weighting of each lens-source pair by the critical
        surface density. A natural choice is -2 which minimizes shape noise.
        Default is -2.
    lens_source_cut : None, float or numpy.ndarray, optional
        Determine the lens-source redshift separation cut. If None, no cut is
        applied. If a float, determines the minimum redshift separation between
        lens and source redshift for lens-source pairs to be used. If an array,
        it has to be the same length as the source table and determines the
        maximum lens redshift for a lens-source pair to be used. Default is 0.
    nside : int, optional
        dsigma uses pixelization to group nearby lenses together and process
        them simultaneously. This parameter determines the number of pixels.
        It has to be a power of 2. May impact performance. Default is 256.
    n_jobs : int, optional
        Number of jobs to run at the same time. Default is 1.
    progress_bar : boolean, option
        Whether to show a progress bar for the main loop over lens pixels.
        Default is False.

    Returns
    -------
    table_l : astropy.table.Table
        Lens catalog with the pre-computation results attached to the table.

    Raises
    ------
    ValueError
        If there are problems in the input.

    """
    try:
        assert cosmology.Ok0 == 0
    except AssertionError:
        raise ValueError('dsigma does not support non-flat cosmologies.')

    if np.any(table_l['z'] < 0):
        raise ValueError('Input lens redshifts must all be non-negative.')
    if not isinstance(nside, int) or not np.isin(nside, 2**np.arange(15)):
        raise ValueError('nside must be a positive power of 2. Received ' +
                         '{}.'.format(nside))
    if not isinstance(n_jobs, int) or n_jobs < 1:
        raise ValueError('Number of jobs must be positive integer. Received ' +
                         '{}.'.format(n_jobs))

    if table_n is not None:
        if 'z_bin' not in table_s.colnames:
            raise ValueError('To use source redshift distributions, the ' +
                             'source table needs to have a `z_bin` column.')
        if not np.issubdtype(table_s['z_bin'].data.dtype, int) or np.amin(
                table_s['z_bin']) < 0:
            raise ValueError('The `z_bin` column in the source table must ' +
                             'contain only non-negative integers.')
        if np.amax(table_s['z_bin']) > table_n['n'].data.shape[1]:
            raise ValueError('The source table contains more redshift bins ' +
                             'than where passed via the nz argument.')

    hp = HEALPix(nside, order='ring')
    pix_l = hp.lonlat_to_healpix(table_l['ra'] * u.deg, table_l['dec'] * u.deg)
    pix_s = hp.lonlat_to_healpix(table_s['ra'] * u.deg, table_s['dec'] * u.deg)
    argsort_pix_l = np.argsort(pix_l)
    argsort_pix_s = np.argsort(pix_s)
    u_pix_l, n_pix_l = np.unique(pix_l, return_counts=True)
    u_pix_l = np.ascontiguousarray(u_pix_l)
    n_pix_l = np.ascontiguousarray(np.cumsum(n_pix_l))
    u_pix_s, n_pix_s = np.unique(pix_s, return_counts=True)
    u_pix_s = np.ascontiguousarray(u_pix_s)
    n_pix_s = np.ascontiguousarray(np.cumsum(n_pix_s))

    table_engine_l = {}
    table_engine_s = {}

    table_engine_l['z'] = np.ascontiguousarray(
        table_l['z'][argsort_pix_l], dtype=np.float64)
    table_engine_s['z'] = np.ascontiguousarray(
        table_s['z'][argsort_pix_s], dtype=np.float64)

    for f, f_name in zip([np.sin, np.cos], ['sin', 'cos']):
        for table, argsort_pix, table_engine in zip(
                [table_l, table_s], [argsort_pix_l, argsort_pix_s],
                [table_engine_l, table_engine_s]):
            for angle in ['ra', 'dec']:
                table_engine['{} {}'.format(f_name, angle)] =\
                    np.ascontiguousarray(f(np.deg2rad(table[angle]))[
                        argsort_pix])

    for key in ['w', 'e_1', 'e_2', 'm', 'e_rms', 'R_2', 'R_11', 'R_22',
                'R_12', 'R_21']:
        if key in table_s.colnames:
            table_engine_s[key] = np.ascontiguousarray(
                table_s[key][argsort_pix_s], dtype=np.float64)

    if lens_source_cut is None:
        z_l_max = np.repeat(np.amax(table_l['z']) + 1, len(table_s))
    elif isinstance(lens_source_cut, numbers.Number):
        z_l_max = table_s['z'] - lens_source_cut
    else:
        z_l_max = np.array(lens_source_cut)

    table_engine_s['z_l_max'] = np.ascontiguousarray(
        z_l_max[argsort_pix_s], dtype=np.float64)

    for table, argsort_pix, table_engine in zip(
            [table_l, table_s], [argsort_pix_l, argsort_pix_s],
            [table_engine_l, table_engine_s]):

        z_min = np.amin(table['z'])
        z_max = np.amax(table['z'])
        z_interp = np.linspace(
            z_min, z_max, max(10, int((z_max - z_min) / 0.0001)))

        table_engine['d_com'] = np.ascontiguousarray(interp1d(
            z_interp, cosmology.comoving_transverse_distance(z_interp).to(
                u.Mpc).value)(table['z'])[argsort_pix])

    if table_c is not None and table_n is None:
        z_min = np.amin(table_l['z'])
        z_max = np.amax(table_l['z'])
        z_interp = np.linspace(
            z_min, z_max, max(10, int((z_max - z_min) / 0.001)))
        f_bias_interp = photo_z_dilution_factor(
            z_interp, table_c, cosmology, weighting=weighting,
            lens_source_cut=lens_source_cut)
        f_bias_interp = interp1d(
            z_interp, f_bias_interp, kind='cubic', bounds_error=False,
            fill_value=(f_bias_interp[0], f_bias_interp[-1]))
        table_engine_l['f_bias'] = np.ascontiguousarray(
            f_bias_interp(np.array(table_engine_l['z'])), dtype=np.float64)
        dz_s_interp = mean_photo_z_offset(
            z_interp, table_c=table_c, cosmology=cosmology,
            weighting=weighting)
        dz_s_interp = interp1d(
            z_interp, dz_s_interp, kind='cubic', bounds_error=False,
            fill_value=(dz_s_interp[0], dz_s_interp[-1]))
        table_engine_l['delta z_s'] = np.ascontiguousarray(
            dz_s_interp(np.array(table_engine_l['z'])), dtype=np.float64)

    elif table_c is None and table_n is not None:
        n_bins = table_n['n'].data.shape[1]
        sigma_crit_eff = np.zeros(len(table_l) * n_bins, dtype=np.float64)
        for i in range(n_bins):
            z_min = np.amin(table_l['z'])
            z_max = min(np.amax(table_l['z']),
                        np.amax(table_n['z'][table_n['n'][:, i] > 0]))
            z_interp = np.linspace(
                z_min, z_max, max(10, int((z_max - z_min) / 0.001)))

            sigma_crit_eff_inv_interp = effective_critical_surface_density(
                z_interp, table_n['z'], table_n['n'][:, i],
                cosmology=cosmology, comoving=comoving)**-1
            sigma_crit_eff_inv_interp = interp1d(
                z_interp, sigma_crit_eff_inv_interp, kind='cubic',
                bounds_error=False,
                fill_value=(sigma_crit_eff_inv_interp[0],
                            sigma_crit_eff_inv_interp[-1]))
            sigma_crit_eff_inv_interp = sigma_crit_eff_inv_interp(
                np.array(table_engine_l['z']))
            sigma_crit_eff_interp = np.repeat(np.inf, len(table_l))
            mask = sigma_crit_eff_inv_interp == 0
            sigma_crit_eff_interp[~mask] = sigma_crit_eff_inv_interp[~mask]**-1
            sigma_crit_eff[i::n_bins] = sigma_crit_eff_interp
        table_engine_l['sigma_crit_eff'] = np.ascontiguousarray(
            sigma_crit_eff, dtype=np.float64)
        table_engine_s['z_bin'] = np.ascontiguousarray(
            table_s['z_bin'][argsort_pix_s], dtype=int)
    elif table_c is not None and table_s is not None:
        raise ValueError('table_c and table_n cannot both be given.')

    # Create arrays that will hold the final results.
    table_engine_r = {}
    n_results = len(table_l) * (len(bins) - 1)

    key_list = ['sum 1', 'sum w_ls', 'sum w_ls e_t', 'sum w_ls z_s',
                'sum w_ls e_t sigma_crit']

    if 'm' in table_s.colnames:
        key_list.append('sum w_ls m')

    if 'e_rms' in table_s.colnames:
        key_list.append('sum w_ls (1 - e_rms^2)')

    if 'R_2' in table_s.colnames:
        key_list.append('sum w_ls A p(R_2=0.3)')

    if (('R_11' in table_s.colnames) and ('R_12' in table_s.colnames) and
            ('R_21' in table_s.colnames) and ('R_22' in table_s.colnames)):
        key_list.append('sum w_ls R_T')

    for key in key_list:
        table_engine_r[key] = np.ascontiguousarray(
            np.zeros(n_results, dtype=(
                np.int64 if key == 'sum 1' else np.float64)))

    z_l = np.array(table_engine_l['z'])
    d_com_l = np.array(table_engine_l['d_com'])

    if not isinstance(bins, u.quantity.Quantity):
        bins = bins * u.Mpc

    try:
        theta_bins = np.tile(bins.to(u.rad).value, len(table_l))
    except UnitConversionError:
        theta_bins = (np.tile(bins.to(u.Mpc).value, len(table_l)) /
                      np.repeat(d_com_l, len(bins))).flatten()
        if not comoving:
            theta_bins *= (1 + np.repeat(z_l, len(bins))).flatten()

    dist_3d_sq_bins = np.minimum(4 * np.sin(theta_bins / 2.0)**2, 2.0)

    # When running in parrallel, replace numpy arrays with shared-memory
    # multiprocessing arrays.
    if n_jobs > 1:
        dist_3d_sq_bins = get_raw_multiprocessing_array(dist_3d_sq_bins)
        for table_engine in [table_engine_l, table_engine_s, table_engine_r]:
            for key in table_engine.keys():
                table_engine[key] = get_raw_multiprocessing_array(
                    table_engine[key])

    # Create a queue that holds all the pixels containing lenses.
    if n_jobs == 1:
        queue = Queue.Queue()
    else:
        queue = mp.Queue()

    for i in range(len(u_pix_l)):
        queue.put(i)

    args = (u_pix_l, n_pix_l, u_pix_s, n_pix_s, dist_3d_sq_bins,
            table_engine_l, table_engine_s, table_engine_r, bins, comoving,
            weighting, nside, queue, progress_bar)

    if n_jobs == 1:
        precompute_engine(*args)
    else:
        processes = []
        for i in range(n_jobs):
            process = mp.Process(target=precompute_engine, args=(*args, ))
            if i == 0:
                args = list(args)
                args[-1] = False
                args = tuple(args)
            process.start()
            processes.append(process)
        for i in range(n_jobs):
            processes[i].join()

    inv_argsort_pix_l = np.argsort(argsort_pix_l)
    for key in table_engine_r.keys():
        table_l[key] = np.array(table_engine_r[key]).reshape(
            len(table_l), len(bins) - 1)[inv_argsort_pix_l]

    table_l['sum w_ls z_l'] = table_l['z'][:, np.newaxis] * table_l['sum w_ls']

    if 'f_bias' in table_engine_l.keys():
        table_l['sum w_ls e_t sigma_crit f_bias'] = (
            np.array(table_engine_l['f_bias'])[inv_argsort_pix_l][
                :, np.newaxis] * table_l['sum w_ls e_t sigma_crit'])

    if 'delta z_s' in table_engine_l.keys():
        table_l['sum w_ls (z_s - delta z_s)'] = (
            table_l['sum w_ls z_s'] - table_l['sum w_ls'] * np.array(
                table_engine_l['delta z_s'])[inv_argsort_pix_l][:, np.newaxis])

    table_l.meta['bins'] = bins
    table_l.meta['comoving'] = comoving
    table_l.meta['H0'] = cosmology.H0.value
    table_l.meta['Ok0'] = cosmology.Ok0
    table_l.meta['Om0'] = cosmology.Om0
    table_l.meta['weighting'] = weighting

    return table_l



"""Physics functions for the dsigma pipeline."""

import numpy as np
from astropy import constants as c
from astropy import units as u
from scipy.special import jv, jn_zeros
from astropy.cosmology import FlatLambdaCDM


__all__ = ['mpc_per_degree', 'projection_angle', 'critical_surface_density',
           'effective_critical_surface_density',
           'lens_magnification_shear_bias']

_sigma_crit_factor = (c.c**2 / (4 * np.pi * c.G)).to(u.Msun / u.pc).value


def mpc_per_degree(z, cosmology=FlatLambdaCDM(H0=100, Om0=0.3),
                   comoving=False):
    """Estimate the angular scale in Mpc/degree at certain redshift.

    Parameters
    ----------
    cosmology : astropy.cosmology, optional
        Cosmology to assume for calculations.
    z : float or numpy.ndarray
        Redshift of the object.
    comoving : boolen
        Use comoving distance instead of physical distance when True.
        Default: False

    Returns
    -------
    float or numpy.ndarray
        Physical scale in unit of Mpc/degree.

    """
    if comoving:
        return (cosmology.comoving_transverse_distance(z).to(u.Mpc).value *
                np.deg2rad(1))

    return (cosmology.angular_diameter_distance(z).to(u.Mpc).value *
            np.deg2rad(1))


def projection_angle(ra_l, dec_l, ra_s, dec_s):
    r"""Calculate projection angle between lens and sources.

    Parameters
    ----------
    ra_l, dec_l : float or numpy.ndarray
        Coordinates of the lens galaxies in degrees.
    ra_s, dec_s : float or numpy.ndarray
        Coordinates of the source galaxies in degrees.

    Returns
    -------
    cos_2phi, sin_2phi : float or numpy.ndarray
        The :math:`\cos` and :math:`\sin` of :math:`2 \phi`, where
        :math:`\phi` is the angle measured from right ascension direction to a
        line connecting the lens and source galaxies.

    """
    # Convert everything into radians.
    ra_l, dec_l = np.deg2rad(ra_l), np.deg2rad(dec_l)
    ra_s, dec_s = np.deg2rad(ra_s), np.deg2rad(dec_s)

    # Calculate the tan(phi).
    mask = np.cos(dec_s) * np.sin(ra_s - ra_l) != 0

    if hasattr(mask, "__len__"):
        tan_phi = (
            (np.cos(dec_l) * np.sin(dec_s) - np.sin(dec_l) * np.cos(dec_s) *
             np.cos(ra_s - ra_l))[mask] /
            (np.cos(dec_s) * np.sin(ra_s - ra_l))[mask])

        cos_2phi = np.repeat(-1.0, len(mask))
        sin_2phi = np.repeat(0.0, len(mask))

        cos_2phi[mask] = (2.0 / (1.0 + tan_phi * tan_phi)) - 1.0
        sin_2phi[mask] = 2.0 * tan_phi / (1.0 + tan_phi * tan_phi)
    elif mask:
        tan_phi = (
            (np.cos(dec_l) * np.sin(dec_s) - np.sin(dec_l) * np.cos(dec_s) *
             np.cos(ra_s - ra_l)) / (np.cos(dec_s) * np.sin(ra_s - ra_l)))

        cos_2phi = (2.0 / (1.0 + tan_phi * tan_phi)) - 1.0
        sin_2phi = (2.0 * tan_phi / (1.0 + tan_phi * tan_phi))
    else:
        cos_2phi = -1
        sin_2phi = 0

    return cos_2phi, sin_2phi


def critical_surface_density(z_l, z_s, cosmology=None, comoving=True, d_l=None,
                             d_s=None):
    """Compute the critical surface density.

    Parameters
    ----------
    z_l : float or numpy.ndarray
        Redshift of lens.
    z_s : float or numpy.ndarray
        Redshift of source.
    cosmology : astropy.cosmology, optional
        Cosmology to assume for calculations. Only used if comoving distances
        are not passed.
    comoving : boolean, optional
        Flag for using comoving instead of physical units.
    d_l : float or numpy.ndarray
        Comoving transverse distance to the lens. If not given, it is
        calculated from the redshift provided.
    d_s : float or numpy.ndarray
        Comoving transverse distance to the source. If not given, it is
        calculated from the redshift provided.

    Returns
    -------
    sigma_crit : float or numpy.ndarray
        Critical surface density for each lens-source pair.

    """
    if d_l is None:
        d_l = cosmology.comoving_transverse_distance(z_l).to(u.Mpc).value
    if d_s is None:
        d_s = cosmology.comoving_transverse_distance(z_s).to(u.Mpc).value

    dist_term = (1e-6 * (d_s / (1 + z_s)) / (d_l / (1 + z_l)) /
                 (np.where(d_s > d_l, d_s - d_l, 1) / (1 + z_s)))

    if np.isscalar(dist_term):
        if d_s <= d_l:
            dist_term = np.inf
    else:
        dist_term[d_s <= d_l] = np.inf

    if comoving:
        dist_term /= (1.0 + z_l)**2

    return _sigma_crit_factor * dist_term


def effective_critical_surface_density(z_l, z_s, n_s, cosmology,
                                       comoving=True):
    """Compute the effective critical surface density.

    Parameters
    ----------
    z_l : float or numpy.ndarray
        Redshift of lens.
    z_s : numpy.ndarray
        Redshifts of sources.
    n_s : numpy.ndarray
        Fraction of source galaxies in each redshift bin. Does not need to be
        normalized.
    cosmology : astropy.cosmology
        Cosmology to assume for calculations.
    comoving : boolean, optional
        Flag for using comoving instead of physical unit.

    Returns
    -------
    sigma_crit_eff : float or numpy.ndarray
        Effective critical surface density for the lens redshift given the
        source redshift distribution.

    """
    d_l = cosmology.comoving_transverse_distance(z_l).to(u.Mpc).value
    d_s = cosmology.comoving_transverse_distance(z_s).to(u.Mpc).value

    if not np.isscalar(z_l):
        z_l = np.repeat(z_l, len(z_s)).reshape((len(z_l), len(z_s)))
        d_l = np.repeat(d_l, len(z_s)).reshape(z_l.shape)
        z_s = np.tile(z_s, len(z_l)).reshape(z_l.shape)
        d_s = np.tile(d_s, len(z_l)).reshape(z_l.shape)
        n_s = np.tile(n_s, len(z_l)).reshape(z_l.shape)

    sigma_crit = critical_surface_density(z_l, z_s, cosmology=cosmology,
                                          comoving=comoving, d_l=d_l, d_s=d_s)

    if not np.isscalar(z_l):
        sigma_crit_eff = np.repeat(np.inf, len(z_l))
        mask = np.average(sigma_crit**-1, axis=-1, weights=n_s) == 0
        sigma_crit_eff[~mask] = np.average(sigma_crit**-1, axis=-1,
                                           weights=n_s)[~mask]**-1
        return sigma_crit_eff
    else:
        if np.average(sigma_crit**-1, weights=n_s) > 0:
            return np.average(sigma_crit**-1, weights=n_s)**-1
        else:
            return np.inf


def lens_magnification_shear_bias(theta, alpha_l, z_l, z_s, camb_results,
                                  n_z=10, n_ell=200, bessel_function_zeros=100,
                                  k_max=1e3):
    r"""Compute the lens magnification bias to the mean tangential shear.

    This function is based on equations (13) and (14) in Unruh et al. (2020).

    Parameters
    ----------
    theta : float
        Angular separation :math:`\theta` from the lens sample in radians.
    alpha_l : float
        Local slope of the flux distribution of lenses near the flux limit.
    z_l : float
        Redshift of lens.
    z_s : float
        Redshift of source.
    camb_results : camb.results.CAMBdata
        CAMB results object that contains information on cosmology and the
        matter power spectrum.
    n_z : int, optional
        Number of redshift bins used in the integral. Larger numbers will be
        more accurate.
    n_ell : int, optional
        Number of :math:`\ell` bins used in the integral. Larger numbers will
        be more accurate.
    bessel_function_zeros : int, optional
        The calculation involves an integral over the second order Bessel
        function :math:`J_2 (\ell \theta)` from :math:`\ell = 0` to
        :math:`\ell = \infty`. In practice, this function replaces the upper
        bound with the bessel_function_zeros-th zero point of the Bessel
        function. Larger number should lead to more accurate results. However,
        in practice, this also requires larger `n_ell`. Particularly, `n_ell`
        should never fall below `bessel_function_zeros`.
    k_max : float, optional
        The maximum wavenumber beyond which the power spectrum is assumed to be
        0.

    Returns
    -------
    et_lm : float
        Bias in the mean tangential shear due to lens magnification effects.

    """
    camb_interp = camb_results.get_matter_power_interpolator(
        hubble_units=False, k_hunit=False)

    ell_min = 0
    ell_max = np.amax(jn_zeros(2, bessel_function_zeros)) / theta
    z_min = 0
    z_max = min(z_l, z_s)

    z, w_z = np.polynomial.legendre.leggauss(n_z)
    z = (z_max - z_min) / 2.0 * z + (z_max + z_min) / 2.0
    w_z = w_z * (z_max - z_min) / 2.0

    ell, w_ell = np.polynomial.legendre.leggauss(n_ell)
    ell = (ell_max - ell_min) / 2.0 * ell + (ell_max + ell_min) / 2.0
    w_ell = w_ell * (ell_max - ell_min) / 2.0

    int_z = np.array([
        (1 + z_i)**2 / (2 * np.pi) *
        camb_results.hubble_parameter(0) /
        camb_results.hubble_parameter(z_i) *
        camb_results.angular_diameter_distance2(z_i, z_l) *
        camb_results.angular_diameter_distance2(z_i, z_s) /
        camb_results.angular_diameter_distance(z_l) /
        camb_results.angular_diameter_distance(z_s) for z_i in z])

    d_ang = np.array([
        camb_results.angular_diameter_distance(z_i) for z_i in z])
    z = np.tile(z, n_ell)
    int_z = np.tile(int_z, n_ell)
    d_ang = np.tile(d_ang, n_ell)
    w_z = np.tile(w_z, n_ell)

    int_ell = ell * jv(2, ell * theta)
    ell = np.repeat(ell, n_z)
    int_ell = np.repeat(int_ell, n_z)
    w_ell = np.repeat(w_ell, n_z)

    k = (ell + 0.5) / ((1 + z) * d_ang)

    int_z_ell = np.array([camb_interp.P(z[i], k[i]) for i in range(len(k))])
    int_z_ell = np.where(k > k_max, 0, int_z_ell)

    gamma = np.sum(int_z * int_ell * int_z_ell * w_z * w_ell)
    gamma = ((gamma * u.Mpc**3) * 9 * camb_results.Params.H0**3 * u.km**3 /
             u.s**3 / u.Mpc**3 *
             (camb_results.Params.omch2 + camb_results.Params.ombh2)**2 /
             (camb_results.Params.H0 / 100)**4 / 4 / c.c**3)

    return 2 * (alpha_l - 1) * gamma



"""Module with functions specific to the CFHTLenS survey."""

__all__ = ['default_version', 'known_versions', 'e_2_convention',
           'default_column_keys']

e_2_convention = 'standard'
default_version = None
known_versions = [None, ]


def default_column_keys(version=None):
    """Return a dictionary of default column keys.

    Parameters
    ----------
    version : optional
        Ignored.

    Returns
    -------
    keys : dict
        Dictionary of default column keys.

    """
    keys = {
        'ra': 'ALPHA_J2000',
        'dec': 'DELTA_J2000',
        'z': 'Z_B',
        'z_low': 'Z_B_MIN',
        'e_1': 'e1',
        'e_2': 'e2',
        'w': 'weight',
        'm': 'm'}
    return keys



"""Module with functions specific to the Kilo Degree Survey."""

import numpy as np

__all__ = ['default_version', 'known_versions', 'e_2_convention',
           'default_column_keys', 'tomographic_redshift_bin',
           'multiplicative_shear_bias']

default_version = 'DR4'
known_versions = ['DR3', 'KV450', 'DR4']
e_2_convention = 'standard'


def default_column_keys(version=default_version):
    """Return a dictionary of default column keys.

    Parameters
    ----------
    version : string or None, optional
        Version of the catalog.

    Returns
    -------
    keys : dict
        Dictionary of default column keys.

    Raises
    ------
    ValueError
        If `version` does not correspond to a known catalog version.

    """
    if version == 'DR3':
        keys = {
            'ra': 'RAJ2000',
            'dec': 'DECJ2000',
            'z': 'Z_B',
            'z_low': 'Z_B_MIN',
            'e_1': 'e1',
            'e_2': 'e2',
            'w': 'weight',
            'm': 'm'}
    elif version == 'KV450':
        keys = {
            'ra': 'ALPHA_J2000',
            'dec': 'DELTA_J2000',
            'z': 'Z_B',
            'z_low': 'Z_B_MIN',
            'e_1': 'bias_corrected_e1',
            'e_2': 'bias_corrected_e2',
            'w': 'weight'}
    elif version == 'DR4':
        keys = {
            'ra': 'ALPHA_J2000',
            'dec': 'DELTA_J2000',
            'z': 'Z_B',
            'z_low': 'Z_B_MIN',
            'e_1': 'e1',
            'e_2': 'e2',
            'w': 'weight'}
    else:
        raise ValueError(
            "Unkown version of KiDS. Supported versions are {}.".format(
                known_versions))

    return keys


def tomographic_redshift_bin(z_s, version=default_version):
    """Return the photometric redshift bin.

    Parameters
    ----------
    z_s : numpy.ndarray
        Photometric redshifts.
    version : string, optional
        Which catalog version to use. Currently ignored.

    Returns
    -------
    z_bin : numpy.ndarray
        The tomographic redshift bin corresponding to each photometric
        redshift. Returns -1 in case a redshift does not fall into any bin.

    """
    z_bin = np.digitize(z_s, [0.1, 0.3, 0.5, 0.7, 0.9, 1.2]) - 1
    z_bin = np.where((z_s < 0.1) | (z_s >= 1.2), -1, z_bin)

    return z_bin


def multiplicative_shear_bias(z_bin, version=default_version):
    """Return the multiplicative shear bias.

    For many version of KiDS, the multiplicative shear bias is not estimated
    on the basis of individual sources but for broad photometric redshift
    bins. This function returns the multiplicative bias :math:`m` as a function
    of the bin.

    Parameters
    ----------
    z_bin : numpy.ndarray
        Tomographic redshift bin.
    version : string
        Which catalog version to use.

    Returns
    -------
    m : numpy.ndarray
        The multiplicative shear bias corresponding to each tomographic bin.

    Raises
    ------
    ValueError
        If the `version` does not correspond to a known catalog version or
        multiplicative shear biases cannot be defined for this version of the
        catalog.

    """
    if version == 'DR3':

        raise ValueError('For DR3, the multiplicative shear bias is ' +
                         'defined for each object individually.')

    elif version in ['KV450', 'DR4']:

        if version == 'KV450':
            m = np.array([-0.017, -0.008, -0.015, 0.010, 0.006])
        else:
            m = np.array([-0.009, -0.011, -0.015, 0.002, 0.007])

        return np.where(z_bin != -1, m[z_bin], np.nan)

    else:
        raise ValueError(
            "Unkown version of KiDS. Supported versions are {}.".format(
                known_versions))



from setuptools import setup, Extension
from Cython.Build import cythonize

setup(
    ext_modules=cythonize([Extension(
        'dsigma.precompute_engine', ['dsigma/precompute_engine.pyx'],
        extra_compile_args=['-O3', '-march=native'])])
)



"""Convenience functions for the dsigma pipeline."""

import numbers
import numpy as np
from astropy.table import Table

from . import surveys

__all__ = ['dsigma_table', 'spherical_to_cartesian', 'cartesian_to_spherical']


def dsigma_table(table, table_type, survey=None, version=None, copy=False,
                 verbose=True, e_2_convention=None, **kwargs):
    """Convert a table into a table accepted by ``dsigma``.

    Convenience function to convert a table into an astropy table that is
    easily parsed by ``dsigma``. Specifically, this table will have all
    necessary columns for ``dsigma`` to work.

    Parameters
    ----------
    table : object
        Input table. This can be any object easily converted into an astropy
        table. For example, a structured numpy array, a list of dictionaries
        or an astropy table all work.
    table_type : string
        String describing the table type. Valid choices are 'lens', 'source'
        or 'calibration'.
    survey : string, optional
        String describing the specific survey. Providing the survey makes the
        function look for specific keys in the input table. Default is None.
    version : string, optional
        If a survey is provided, one can specify the version or data release
        of the survey. If None, it will assume the most recent public vercsion.
        Default is None.
    copy : boolean, optional
        Whether the output table shares memory with the input table. Setting to
        False can save memory. However, data will be corupted if the original
        input table is manipulated. Default is False.
    verbose : boolean, optional
        Whether to output information about the assignments. Default is True.
    e_2_convention : string, optional
        Whether to switch the sign of e_2 in the input catalog. If 'standard',
        e_2 is not changed. On the other hand, if 'flipped', the sign of e_2
        will be changed. If None, it defaults to 'standard' unless overwritten
        by the specific survey. Default is None.
    **kwargs : dict, optional
        This function has a set of default keys for columns in the input table
        that it associates with certain data. These default choices might be
        overwritten by survey-specific keys. If that does not work
        sufficiently, keys can be overwritten by the user with this argument.
        For example, ``ra='R.A.'`` would associate the right ascensions with
        the ``R.A.`` key in the input table. Alternatively, one can also
        provide numbers that will be applied to all entries. This only really
        makes sense for weights that one wants to ignore, i.e. ``w_sys=1``.

    Returns
    -------
    table_out : astropy.table.Table
        Table with all necessary data to be parsed by ``dsigma``.

    Raises
    ------
    ValueError
        If `table_type`, `survey` or `e_2_convention` are not valid options.
    RuntimeError
        If the input table contains masked data.

    """

    table = Table(table, copy=copy)

    if table_type not in ['lens', 'source', 'calibration']:
        raise ValueError(
            "The table_type argument must be 'lens', 'source' or " +
            "'calibration' but received '{}'.".format(table_type))

    if (survey is not None and survey.lower().split(':')[0] not in
            surveys.__all__):
        raise ValueError('Unknown survey. The known surveys are: {}.'.format(
                         ', '.join(surveys.__all__)))

    if (e_2_convention is not None and e_2_convention not in
            ['standard', 'flipped']):
        raise ValueError("e_2_convention must be None, 'standard' or " +
                         "flipped but received '{}'".format(e_2_convention))

    # Set the generic keys.
    keys = {'z': 'z'}

    if table_type != 'calibration':
        keys['ra'] = 'ra'
        keys['dec'] = 'dec'
    if table_type != 'source':
        keys['w_sys'] = 'w_sys'
    if table_type != 'lens':
        keys['w'] = 'w'
    if table_type == 'source':
        keys['e_1'] = 'e_1'
        keys['e_2'] = 'e_2'
    if table_type == 'calibration':
        keys['z_true'] = 'z_true'

    # Overwrite with survey-specific keys.
    if survey is not None:
        table.meta['survey'] = survey

        if version is None:
            version = getattr(surveys, survey.lower()).default_version
        table.meta['version'] = version
        keys.update(getattr(surveys, survey.lower()).default_column_keys(
            version=version))
        if table_type == 'calibration':
            del keys['ra']
            del keys['dec']
            del keys['e_1']
            del keys['e_2']
            del keys['w']
            del keys['m']
            keys.pop('e_rms', None)
            keys.pop('R_2', None)
        if e_2_convention is None:
            e_2_convention = getattr(surveys, survey.lower()).e_2_convention

    # Finally, update with user-specified keys.
    keys.update(**kwargs)

    if verbose:
        print("Assignment for {} table...".format(table_type))
        for output_key, input_key in keys.items():
            print("    {:<10} -> {}".format(output_key, input_key))

    # Assert columns exist, re-name them and drop unnecessary columns.
    for input_key in keys.values():
        try:
            assert input_key in table.colnames or isinstance(input_key,
                                                             numbers.Number)
        except AssertionError:
            raise ValueError("Key '{}' must be present in ".format(input_key) +
                             "the input table.")

    # Keep only those columns with relevant data.
    table_out = Table()

    for output_key, input_key in keys.items():
        if isinstance(input_key, numbers.Number):
            table_out[output_key] = np.repeat(input_key, len(table))
        else:
            if isinstance(table[input_key].data, np.ma.MaskedArray):
                if np.ma.is_masked(table[input_key].data):
                    raise RuntimeError('Input table contained masked data.')
                table_out[output_key] = table[input_key].data.filled()
            else:
                table_out[output_key] = table[input_key].data

    if e_2_convention == 'flipped' and table_type == 'source':
        table_out['e_2'] = - table_out['e_2']
        if verbose:
            print("Info: Flipping sign of e_2 component.")

    return table_out


def spherical_to_cartesian(ra, dec):
    """Convert spherical coordinates to cartesian coordinates on a unit sphere.

    Parameters
    ----------
    ra : float or numpy.ndarray
        Right ascension.
    dec : float or numpy.ndarray
        Declination.

    Returns
    -------
    x, y, z : float or numpy.ndarray
        Cartesian coordinates.

    """
    x = np.cos(np.deg2rad(ra)) * np.cos(np.deg2rad(dec))
    y = np.sin(np.deg2rad(ra)) * np.cos(np.deg2rad(dec))
    z = np.sin(np.deg2rad(dec))
    return x, y, z


def cartesian_to_spherical(x, y, z):
    """Convert cartesian coordinates to spherical coordinates.

    Parameters
    ----------
    x : float or numpy.ndarray
        x-coordinate.
    y : float or numpy.ndarray
        y-coordinate.
    z : float or numpy.ndarray
        z-coordinate.

    Returns
    -------
    ra, dec : float or numpy.ndarray
        Spherical coordinates.

    """
    r = np.sqrt(x**2 + y**2 + z**2)
    ra = np.arctan2(y, x)
    ra = np.where(ra < 0, ra + 2 * np.pi, ra)
    dec = np.arcsin(z / r)
    return np.rad2deg(ra), np.rad2deg(dec)



"""Module with functions specific to the Dark Energy Survey."""

import numpy as np

__all__ = ['default_version', 'known_versions', 'e_2_convention',
           'default_column_keys', 'tomographic_redshift_bin',
           'multiplicative_shear_bias', 'selection_response']

default_version = 'Y3'
known_versions = ['Y1', 'Y3']
e_2_convention = 'standard'


def default_column_keys(version=default_version):
    """Return a dictionary of default column keys.

    Parameters
    ----------
    version : string or None, optional
        Version of the catalog.

    Returns
    -------
    keys : dict
        Dictionary of default column keys.

    Raises
    ------
    ValueError
        If `version` does not correspond to a known catalog version.

    """
    if version == 'Y1':
        keys = {
            'ra': 'ra',
            'dec': 'dec',
            'z': 'MEAN_Z',
            'z_true': 'Z_MC',
            'e_1': 'e1',
            'e_2': 'e2',
            'w': 1,
            'R_11': 'R11',
            'R_12': 'R12',
            'R_21': 'R21',
            'R_22': 'R22',
            'flags_select': 'flags_select',
            'flags_select_1p': 'flags_select_1p',
            'flags_select_1m': 'flags_select_1m',
            'flags_select_2p': 'flags_select_2p',
            'flags_select_2m': 'flags_select_2m'}
    elif version == 'Y3':
        keys = {
            'ra': 'ra',
            'dec': 'dec',
            'z': np.nan,
            'z_bin': 'bhat',
            'e_1': 'e_1',
            'e_2': 'e_2',
            'w': 'weight',
            'w_1p': 'weight_1p',
            'w_1m': 'weight_1m',
            'w_2p': 'weight_2p',
            'w_2m': 'weight_2m',
            'R_11': 'R11',
            'R_12': 'R12',
            'R_21': 'R21',
            'R_22': 'R22',
            'flags_select': 'flags_select',
            'flags_select_1p': 'flags_select_1p',
            'flags_select_1m': 'flags_select_1m',
            'flags_select_2p': 'flags_select_2p',
            'flags_select_2m': 'flags_select_2m'}
    else:
        raise ValueError(
            "Unkown version of DES. Supported versions are {}.".format(
                known_versions))

    return keys


def tomographic_redshift_bin(z_s, version=default_version):
    """Return the photometric redshift bin.

    Parameters
    ----------
    z_s : numpy.ndarray
        Photometric redshifts.
    version : string, optional
        Which catalog version to use.

    Returns
    -------
    z_bin : numpy.ndarray
        The tomographic redshift bin corresponding to each photometric
        redshift. Returns -1 in case a redshift does not fall into any bin.

    Raises
    ------
    ValueError
        If the `version` does not correspond to a known catalog version or
        if tomographic bins were not assigned based on photometric redshifts
        for the given catalog version.

    """
    if version == 'Y1':
        z_bins = [0.2, 0.43, 0.63, 0.9, 1.3]
    elif version == 'Y3':
        raise ValueError('DES Y3 assigns redshift bins based on colors, ' +
                         'not photometric redshifts.')
    else:
        raise ValueError(
            "Unkown version of DES. Supported versions are {}.".format(
                known_versions))

    z_bin = np.digitize(z_s, z_bins) - 1
    z_bin = np.where((z_s < np.amin(z_bins)) | (z_s >= np.amax(z_bins)) |
                     np.isnan(z_s), -1, z_bin)

    return z_bin


def multiplicative_shear_bias(z_bin, version=default_version):
    """Return the multiplicative shear bias.

    For DES Y3, we can define a blending-related multiplicative shear bias.
    This function returns the multiplicative bias :math:`m` as a function
    of the bin. The values can be computed from the blending-corrected Y3
    redshift distributions and are, as expected, very similar to the values in
    Table 4 in MacCrann et al. (2022) where they were calculated for mock
    catalogs.

    Parameters
    ----------
    z_bin : numpy.ndarray
        Tomographic redshift bin.
    version : string, optional
        Which catalog version to use.

    Returns
    -------
    m : numpy.ndarray
        The multiplicative shear bias corresponding to each tomographic bin.

    Raises
    ------
    ValueError
        If the `version` does not correspond to a known catalog version or
        multiplicative shear biases cannot be defined for this version of the
        catalog.

    """
    if version == 'Y1':
        raise ValueError(
            'For Y1, we cannot define a multiplicative shear bias.')

    elif version == 'Y3':
        m = np.array([-0.63, -1.98, -2.41, -3.69]) * 0.01
        return np.where(z_bin != -1, m[z_bin], np.nan)

    else:
        raise ValueError(
            "Unkown version of DES. Supported versions are {}.".format(
                known_versions))


def selection_response(table_s, version=default_version):
    """Calculate the DES selection response.

    See Sheldon & Huff (2017) and McClintock et al. (2018) for details.

    Parameters
    ----------
    table_s : astropy.table.Table
        Catalog of sources.
    version : string, optional
        Which catalog version to use.

    Returns
    -------
    R_sel : numpy.ndarray
        2x2 matrix containing the selection response.

    """
    R_sel = np.zeros((2, 2))

    for i in range(2):
        for j in range(2):

            if np.issubdtype(table_s['flags_select_1p'].dtype, np.integer):
                select_p = table_s['flags_select_{}p'.format(i + 1)] == 0
                select_m = table_s['flags_select_{}m'.format(i + 1)] == 0
            else:
                select_p = table_s['flags_select_{}p'.format(i + 1)]
                select_m = table_s['flags_select_{}m'.format(i + 1)]
            e = table_s['e_{}'.format(j + 1)]

            if version == 'Y3':
                w_p = table_s['w_{}p'.format(i + 1)]
                w_m = table_s['w_{}m'.format(i + 1)]
            else:
                w_p = np.ones(len(table_s))
                w_m = np.ones(len(table_s))

            R_sel[i, j] = (
                np.average(e[select_p], weights=w_p[select_p]) -
                np.average(e[select_m], weights=w_m[select_m])) / 0.02

    return R_sel
